//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/ICoordSystem.cpp
//! @brief     Implements ICoordSystem class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Coord/ICoordSystem.h"
#include "Device/Coord/AxisNames.h"
#include "Device/Data/Powerfield.h"

ICoordSystem::~ICoordSystem() = default;

std::unique_ptr<Powerfield<double>> ICoordSystem::createPowerfield(Axes::Coords units) const
{
    std::unique_ptr<Powerfield<double>> result = std::make_unique<Powerfield<double>>();
    for (size_t i = 0; i < dimension(); ++i)
        result->addAxis(*createConvertedAxis(i, units));
    result->setAllTo(0.0);
    return result;
}

std::string ICoordSystem::axisName(size_t i_axis, const Axes::Coords units) const
{
    const auto& name_maps = createNameMaps();
    ASSERT(i_axis < name_maps.size());
    const auto& name_map = name_maps[i_axis];
    const auto& it = name_map.find(units == Axes::Coords::UNDEFINED ? defaultUnits() : units);
    ASSERT(it != name_map.cend());
    return it->second;
}

std::unique_ptr<Powerfield<double>>
ICoordSystem::createConvertedData(const Powerfield<double>& data, Axes::Coords units) const
{
    std::unique_ptr<Powerfield<double>> result(new Powerfield<double>);
    for (size_t i = 0; i < data.rank(); ++i)
        result->addAxis(*createConvertedAxis(i, units));
    result->setRawDataVector(data.getRawDataVector());
    return result;
}

void ICoordSystem::throwUnitsError(std::string method, std::vector<Axes::Coords> available) const
{
    std::stringstream ss;
    ss << "Unit type error in " << method
       << ": unknown or unsupported unit type. Available units "
          "are:\n";
    for (auto unit : available)
        ss << axisUnitLabel.at(unit) << "\n";
    throw std::runtime_error(ss.str());
}

Axes::Coords ICoordSystem::substituteDefaultUnits(Axes::Coords units) const
{
    return units == Axes::Coords::UNDEFINED ? defaultUnits() : units;
}
