//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/ArrayUtils.cpp
//! @brief     Implements internal functions
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Data/ArrayUtils.h"

std::pair<size_t, size_t> DataUtils::Array::getShape(const std::vector<std::vector<double>>& data)
{
    size_t nrows = data.size();
    size_t ncols(0);
    if (nrows)
        ncols = data[0].size();
    for (size_t row = 0; row < nrows; row++)
        if (data[row].size() != ncols)
            throw std::runtime_error("Invalid data: not rectangular");
    return std::make_pair(nrows, ncols);
}

std::unique_ptr<Powerfield<double>> DataUtils::Array::createData(const std::vector<double>& vec)
{
    auto result = std::make_unique<Powerfield<double>>();
    const size_t length = vec.size();
    result->addAxis(FixedBinAxis("axis0", length, 0.0, static_cast<double>(length)));
    result->setRawDataVector(vec);
    return result;
}

std::unique_ptr<Powerfield<double>>
DataUtils::Array::createData(const std::vector<std::vector<double>>& vec)
{
    auto result = std::make_unique<Powerfield<double>>();

    auto shape = DataUtils::Array::getShape(vec);
    const size_t nrows = shape.first;
    const size_t ncols = shape.second;

    if (nrows == 0 || ncols == 0)
        throw std::runtime_error(
            "Error in DataUtils::Array::createDataImpl: input argument contains empty dimensions");

    result->addAxis(FixedBinAxis("axis0", ncols, 0.0, static_cast<double>(ncols)));
    result->addAxis(FixedBinAxis("axis1", nrows, 0.0, static_cast<double>(nrows)));

    // filling the data
    for (size_t row = 0; row < nrows; ++row) {
        for (size_t col = 0; col < ncols; ++col) {
            size_t globalbin = nrows - row - 1 + col * nrows;
            (*result)[globalbin] = vec[row][col];
        }
    }

    return result;
}

std::vector<double> DataUtils::Array::createVector1D(const Powerfield<double>& data)
{
    if (data.rank() != 1)
        throw std::runtime_error("DataUtils::Array::createVector1D() -> Error. Not 1D data.");

    std::vector<double> result = data.getRawDataVector();
    return result;
}

std::vector<std::vector<double>> DataUtils::Array::createVector2D(const Powerfield<double>& data)
{
    std::vector<std::vector<double>> result;

    const size_t nrows = data.axis(1).size();
    const size_t ncols = data.axis(0).size();

    result.resize(nrows);

    for (size_t row = 0; row < nrows; ++row) {
        result[row].resize(ncols, 0.0);
        for (size_t col = 0; col < ncols; ++col) {
            size_t globalbin = nrows - row - 1 + col * nrows;
            result[row][col] = data[globalbin];
        }
    }

    return result;
}
