//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/IOFactory.cpp
//! @brief     Implements class PowerfieldIOFactory.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/IOFactory.h"
#include "Base/Util/FileSystemUtils.h"
#include "Device/Histo/IHistogram.h"
#include "Device/Histo/SimulationResult.h"
#include "Device/InputOutput/DataFormatUtils.h"
#include "Device/InputOutput/ReadReflectometry.h"
#include "Device/InputOutput/ReadWriteINT.h"
#include "Device/InputOutput/ReadWriteNicos.h"
#include "Device/InputOutput/ReadWriteNumpyTXT.h"
#include "Device/InputOutput/ReadWriteTiff.h"
#include <exception>
#include <fstream>
#include <memory>

#ifdef _WIN32
#pragma warning(push)
#pragma warning(disable : 4244 4275)
#include "Device/InputOutput/boost_streams.h"
#pragma warning(pop)
#else
#include "Device/InputOutput/boost_streams.h"
#endif

Powerfield<double>* IOFactory::readPowerfield(const std::string& file_name, LoaderSelector selector)
{
    const auto readAs = [=](LoaderSelector testForSelector) {
        return (selector == testForSelector)
               || (selector == automatic
                   && fileTypeMatchesLoaderSelector(file_name, testForSelector));
    };

    Powerfield<double>* ret = nullptr;

    if (readAs(bornagain))
        ret = readPowerfield(file_name,
                             [](std::istream& s) { return ReadWriteINT().readPowerfield(s); });

    else if (readAs(nicos))
        ret = readPowerfield(file_name, [](std::istream& s) { return IO::readNicosData(s); });

#ifdef BA_TIFF_SUPPORT
    else if (readAs(tiff))
        ret = readPowerfield(file_name,
                             [](std::istream& s) { return ReadWriteTiff().readPowerfield(s); });
#endif

    else
        // Try to read ASCII by default. Binary maps to ASCII.
        // If the file is not actually a matrix of numbers,
        // the error will be thrown during the reading.
        ret = readPowerfield(file_name,
                             [](std::istream& s) { return ReadWriteNumpyTXT().readPowerfield(s); });

    ASSERT(ret);
    return ret;
}

Powerfield<double>* IOFactory::readReflectometryData(const std::string& file_name)
{
    return readPowerfield(file_name,
                          [](std::istream& s) { return ReadReflectometry().readPowerfield(s); });
}

IHistogram* IOFactory::readIntensityData(const std::string& file_name)
{
    std::unique_ptr<Powerfield<double>> data(readPowerfield(file_name));
    if (!data)
        throw std::runtime_error("Cannot read intensity data from file: " + file_name);
    return IHistogram::createHistogram(*data);
}

void IOFactory::writePowerfield(const Powerfield<double>& data, const std::string& file_name)
{
    if (DataUtils::Format::isIntFile(file_name))
        writePowerfield(file_name,
                        [&](std::ostream& s) { ReadWriteINT().writePowerfield(data, s); });
#ifdef BA_TIFF_SUPPORT
    else if (DataUtils::Format::isTiffFile(file_name))
        writePowerfield(file_name,
                        [&](std::ostream& s) { ReadWriteTiff().writePowerfield(data, s); });
#endif
    else
        writePowerfield(file_name,
                        [&](std::ostream& s) { ReadWriteNumpyTXT().writePowerfield(data, s); });
}

void IOFactory::writePowerfield(const std::string& file_name,
                                std::function<void(std::ostream&)> writeData)
{
    using namespace DataUtils::Format;

    std::ofstream fout;
    std::ios_base::openmode openmode = std::ios::out;
    if (isTiffFile(file_name) || isCompressed(file_name))
        openmode = std::ios::out | std::ios_base::binary;

#ifdef _WIN32
    fout.open(BaseUtils::Filesystem::convert_utf8_to_utf16(file_name), openmode);
#else
    fout.open(file_name, openmode);
#endif

    if (!fout.is_open())
        throw std::runtime_error("Cannot open file for writing: " + file_name);
    if (!fout.good())
        throw std::runtime_error("File is not good, probably it is a directory: " + file_name);
    std::stringstream ss;
    writeData(ss);

    boost::iostreams::filtering_streambuf<boost::iostreams::input> input_filtered;
    if (DataUtils::Format::isGZipped(file_name))
        input_filtered.push(boost::iostreams::gzip_compressor());
    else if (DataUtils::Format::isBZipped(file_name))
        input_filtered.push(boost::iostreams::bzip2_compressor());
    input_filtered.push(ss);

    boost::iostreams::copy(input_filtered, fout);

    fout.close();
}

bool IOFactory::fileTypeMatchesLoaderSelector(const std::string& fileName, LoaderSelector selector)
{
    switch (selector) {
    case bornagain:
        return DataUtils::Format::isIntFile(fileName);
    case nicos:
        return DataUtils::Format::isNicosFile(fileName);
    case tiff:
        return DataUtils::Format::isTiffFile(fileName);
    case automatic:
        return false;
    }

    return false;
}

void IOFactory::writeIntensityData(const IHistogram& histogram, const std::string& file_name)
{
    std::unique_ptr<Powerfield<double>> data(histogram.createPowerfield());
    writePowerfield(*data, file_name);
}

void IOFactory::writeSimulationResult(const SimulationResult& result, const std::string& file_name)
{
    auto data = result.data();
    writePowerfield(*data, file_name);
}

Powerfield<double>*
IOFactory::readPowerfield(const std::string& file_name,
                          std::function<Powerfield<double>*(std::istream&)> readData)
{
    if (!BaseUtils::Filesystem::IsFileExists(file_name))
        throw std::runtime_error("File does not exist: " + file_name);

    using namespace DataUtils::Format;
    std::ifstream input_stream;
    std::ios_base::openmode openmode = std::ios::in;
    if (isTiffFile(file_name) || isCompressed(file_name))
        openmode = std::ios::in | std::ios_base::binary;

#ifdef _WIN32
    input_stream.open(BaseUtils::Filesystem::convert_utf8_to_utf16(file_name), openmode);
#else
    input_stream.open(file_name, openmode);
#endif

    if (!input_stream.is_open())
        throw std::runtime_error("Cannot open file for reading: " + file_name);
    if (!input_stream.good())
        throw std::runtime_error("File is not good, probably it is a directory:" + file_name);

    boost::iostreams::filtering_streambuf<boost::iostreams::input> input_filtered;
    if (DataUtils::Format::isGZipped(file_name))
        input_filtered.push(boost::iostreams::gzip_decompressor());
    else if (DataUtils::Format::isBZipped(file_name))
        input_filtered.push(boost::iostreams::bzip2_decompressor());
    input_filtered.push(input_stream);
    // we use stringstream since it provides random access which is important for tiff files
    std::stringstream str;
    boost::iostreams::copy(input_filtered, str);

    return readData(str);
}
