//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/JobFunctions.cpp
//! @brief     Implements auxiliary functions in JobModelFunctions namespace.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/JobFunctions.h"
#include "Base/Util/Assert.h"
#include "Device/Instrument/Instrument.h"
#include "GUI/Model/Data/Data1DViewItem.h"
#include "GUI/Model/Data/DataPropertyContainer.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Job/MinimizerItem.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Support/IO/ItemFileNameUtils.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/Path.h"

namespace {

//! Links RealDataItem to the JobItem's instrument.
// (re-)Linking is necessary because of following reason
// 1) Copying of RealDataItem from RealDataModel on board of JobItem requires relink to the copied
//    insturment
// 2) During relink all masks (if exists) will be converted to the default units of current detector
void processInstrumentLink(JobItem* jobItem)
{
    RealDataItem* realData = jobItem->realDataItem();
    if (!realData)
        throw Error("GUI::Model::JobFunctions::processInstrumentLink() -> Error. No data.");

    realData->linkToInstrument(jobItem->instrumentItem());
}

//! Copies masks and ROI from RealDataItem on board of instrument.
void copyMasksToInstrument(JobItem* jobItem)
{
    auto* mask_container = jobItem->realDataItem()->maskContainerItem();
    jobItem->instrumentItem()->importMasks(mask_container);
}

//! Crops RealDataItem to the region of interest.
void cropRealData(JobItem* jobItem)
{
    ASSERT(jobItem->instrumentItem()->is<Instrument2DItem>());

    RealDataItem* realData = jobItem->realDataItem();

    // adjusting real data to the size of region of interest
    IntensityDataItem* intensityItem = realData->intensityDataItem();

    std::unique_ptr<Powerfield<double>> origData(intensityItem->getPowerfield()->clone());

    auto* instrument_item = jobItem->instrument2DItem();
    GUI::Model::JobItemUtils::createDefaultDetectorMap(intensityItem, instrument_item);

    instrument_item->createInstrument()->getDetector()->iterateOverNonMaskedPoints(
        [&](IDetector::const_iterator it) {
            auto* cropped_data = intensityItem->getPowerfield();
            (*cropped_data)[it.roiIndex()] = (*origData)[it.detectorIndex()];
        });

    intensityItem->updateDataRange();
}

//! Creates necessary fit containers for jobItem intended for fitting.
void createFitContainers(JobItem* jobItem)
{
    FitSuiteItem* fitSuiteItem = jobItem->createFitSuiteItem();

    fitSuiteItem->createFitParametersContainer();
    fitSuiteItem->createMinimizerContainer();
}

} // namespace

//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

void GUI::Model::JobFunctions::initDataView(JobItem* job_item)
{
    ASSERT(job_item && job_item->isValidForFitting());
    ASSERT(job_item->instrumentItem() && job_item->instrumentItem()->is<SpecularInstrumentItem>());
    ASSERT(!job_item->dataItemView());

    auto* view_item = job_item->createDataViewItem();

    auto* property_container = view_item->createPropertyContainerItem();

    property_container->addItem(job_item->realDataItem()->dataItem());
    property_container->addItem(job_item->dataItem());

    // also triggers Data1DViewItem::setAxesRangeToData and
    // GUI::Model::DataViewUtils::updateAxesTitle by setting new value of P_AXES_UNITS.
    auto* converter = job_item->instrumentItem()->createCoordSystem();
    view_item->setAxesUnits(GUI::Model::JobItemUtils::availableUnits(*converter));
}

void GUI::Model::JobFunctions::setupJobItemInstrument(JobItem* jobItem, const InstrumentItem* from)
{
    auto* copiedInstrument = jobItem->copyInstrumentIntoJob(from);
    copiedInstrument->setInstrumentName(copiedInstrument->instrumentType() + " instrument");
    copiedInstrument->setId(QUuid::createUuid().toString());
    jobItem->setInstrumentName(from->instrumentName());
}

//! Setup items intended for storing results of the job.

void GUI::Model::JobFunctions::setupJobItemOutput(JobItem* jobItem)
{
    const bool isSpecularInstrument = jobItem->instrumentItem()->is<SpecularInstrumentItem>();
    const bool isIntensityInstrument = jobItem->instrumentItem()->is<GISASInstrumentItem>()
                                       || jobItem->instrumentItem()->is<OffSpecularInstrumentItem>()
                                       || jobItem->instrumentItem()->is<DepthProbeInstrumentItem>();

    ASSERT(isSpecularInstrument || isIntensityInstrument);

    if (isSpecularInstrument)
        jobItem->setDataType<SpecularDataItem>();
    else if (isIntensityInstrument)
        jobItem->setDataType<IntensityDataItem>();
    else
        ASSERT(0);
}

//! Setups JobItem for fit.

void GUI::Model::JobFunctions::setupJobItemForFit(JobItem* jobItem,
                                                  const RealDataItem* realDataItem)
{
    ASSERT(jobItem->instrumentItem());

    copyRealDataItem(jobItem, realDataItem);
    processInstrumentLink(jobItem);
    copyMasksToInstrument(jobItem);

    if (jobItem->instrumentItem()->is<GISASInstrumentItem>())
        cropRealData(jobItem);
    if (jobItem->instrumentItem()->is<SpecularInstrumentItem>())
        initDataView(jobItem);

    createFitContainers(jobItem);
}

void GUI::Model::JobFunctions::copyRealDataItem(JobItem* jobItem, const RealDataItem* realDataItem)
{
    if (!realDataItem)
        return;

    RealDataItem* realDataItemCopy = jobItem->copyRealDataIntoJob(realDataItem);
    ASSERT(realDataItemCopy);

    realDataItemCopy->dataItem()->setPowerfield(realDataItem->dataItem()->getPowerfield()->clone());

    // adapting the name to job name
    realDataItemCopy->dataItem()->setFileName(
        GUI::Model::FilenameUtils::jobReferenceFileName(jobItem->jobName()));

    // #baimport ++ copy members of realDataItem?

    if (!realDataItem->hasNativeData())
        return;

    realDataItemCopy->setNativePowerfield(realDataItem->nativePowerfield()->clone());
    realDataItemCopy->nativeData()->setFileName(
        GUI::Model::FilenameUtils::jobNativeDataFileName(jobItem->getIdentifier()));
}
