//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/To/ToDomain.cpp
//! @brief     Implements class TransformToDomain
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/To/ToDomain.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Descriptor/UIntDescriptor.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/LayerRoughnessItems.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Util/Error.h"
#include "Sample/Aggregate/IInterference.h"
#include "Sample/Aggregate/ParticleLayout.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/Particle/IParticle.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"

namespace {

std::unique_ptr<MultiLayer> createMultiLayer(const MultiLayerItem& item)
{
    auto P_multilayer = std::make_unique<MultiLayer>();
    double cross_corr_length = item.crossCorrLength();
    if (cross_corr_length > 0)
        P_multilayer->setCrossCorrLength(cross_corr_length);
    R3 external_field = item.externalField();
    P_multilayer->setExternalField(external_field);
    return P_multilayer;
}

std::unique_ptr<Layer> createLayer(const LayerItem& item)
{
    const bool isFirstOrLastLayer = item.isTopLayer() || item.isBottomLayer();
    auto P_layer = std::make_unique<Layer>(*item.materialItem()->createMaterial(),
                                           isFirstOrLastLayer ? 0.0 : item.thickness());
    P_layer->setNumberOfSlices(item.numSlices());
    return P_layer;
}

std::unique_ptr<ParticleLayout> createParticleLayout(const ParticleLayoutItem& item)
{
    auto P_layout = std::make_unique<ParticleLayout>();
    double total_density = item.totalDensity();
    double layout_weight = item.weight();
    P_layout->setTotalParticleSurfaceDensity(total_density);
    P_layout->setWeight(layout_weight);
    return P_layout;
}

std::unique_ptr<IParticle> createIParticle(const ItemWithParticles& item)
{
    std::unique_ptr<IParticle> P_particle;
    if (const auto* particle_item = dynamic_cast<const ParticleItem*>(&item))
        P_particle = particle_item->createParticle();
    else if (const auto* particle_coreshell_item =
                 dynamic_cast<const ParticleCoreShellItem*>(&item))
        P_particle = particle_coreshell_item->createParticleCoreShell();
    else if (const auto* particle_composition_item =
                 dynamic_cast<const ParticleCompositionItem*>(&item))
        P_particle = particle_composition_item->createParticleComposition();
    else if (const auto* mesocrystal_item = dynamic_cast<const MesoCrystalItem*>(&item))
        P_particle = mesocrystal_item->createMesoCrystal();

    return P_particle;
}

std::unique_ptr<ParticleLayout> buildParticleLayout(const ParticleLayoutItem& item)
{
    auto P_layout = createParticleLayout(item);
    auto particles = item.particles();
    for (auto particle : item.particles()) {
        auto P_particle = createIParticle(*particle);
        if (P_particle) {
            P_layout->addParticle(*P_particle);
            continue;
        }

        throw Error("GUI::Model::DomainObjectBuilder::buildParticleLayout()"
                    " -> Error! Not implemented");
    }
    InterferenceItem* interference = item.interference().currentItem();
    if (interference) {
        auto P_interference = interference->createInterference();
        if (P_interference)
            P_layout->setInterference(*P_interference);
    }
    return P_layout;
}

std::unique_ptr<Layer> buildLayer(const LayerItem& item)
{
    auto P_layer = createLayer(item);
    for (ParticleLayoutItem* layout : item.layouts()) {
        auto P_layout = buildParticleLayout(*layout);
        if (P_layout)
            P_layer->addLayout(*P_layout);
    }
    return P_layer;
}

} // namespace

std::unique_ptr<LayerRoughness>
GUI::Transform::ToDomain::createLayerRoughness(LayerBasicRoughnessItem* roughness)
{
    if (!roughness)
        return nullptr;
    return std::make_unique<LayerRoughness>(roughness->sigma(), roughness->hurst(),
                                            roughness->lateralCorrelationLength());
}

std::unique_ptr<MultiLayer>
GUI::Transform::ToDomain::buildMultiLayer(const MultiLayerItem& multiLayerItem)
{
    auto P_multilayer = createMultiLayer(multiLayerItem);
    for (auto* layerItem : multiLayerItem.layers()) {
        auto P_layer = buildLayer(*layerItem);
        ASSERT(P_layer);
        const auto roughness = layerItem->roughness().currentItem();
        auto P_roughness = createLayerRoughness(roughness);
        if (P_roughness && !layerItem->isTopLayer())
            P_multilayer->addLayerWithTopRoughness(*P_layer, *P_roughness);
        else
            P_multilayer->addLayer(*P_layer);
    }
    return P_multilayer;
}
