//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Realspace/RealSpaceBuilderUtils.h
//! @brief     Defines GUI::RealSpace::BuilderUtils namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_REALSPACE_REALSPACEBUILDERUTILS_H
#define BORNAGAIN_GUI_VIEW_REALSPACE_REALSPACEBUILDERUTILS_H

#include "Sample/Scattering/Rotations.h"
#include <QVector3D>
#include <functional>
#include <memory>
#include <utility>

class SessionItem;
struct SceneGeometry;
class RealSpaceBuilder;
class IRotation;
class MesoCrystalItem;
class Particle;
class ParticleComposition;
class ParticleCoreShell;
class ParticleLayoutItem;
class IInterference;
class Particle3DContainer;
class Interference2DParaCrystal;

namespace GUI::RealSpace {

struct Vector3D;
class Model;
namespace Particles {

class Particle;
}

} // namespace GUI::RealSpace

namespace GUI::RealSpace {

class BuilderUtils {
public:
    BuilderUtils(std::function<QColor(const QString&)> fnColorFromMaterialName);

    // compute cumulative abundances of particles
    QVector<double> computeCumulativeAbundances(const ParticleLayoutItem& layoutItem);

    void populateParticlesAtLatticePositions(
        const std::vector<std::vector<double>>& lattice_positions,
        const std::vector<Particle3DContainer>& particle3DContainer_vector,
        GUI::RealSpace::Model* model, const SceneGeometry& sceneGeometry,
        const RealSpaceBuilder* builder3D);

    // Implement Rotation of a 3D particle using parameters from IRotation Object
    GUI::RealSpace::Vector3D implementParticleRotationfromIRotation(const IRotation*& rotation);

    //! Apply transformations (translation, rotation) to a 3D Particle
    //! or to a particle belonging to a ParticleComposition
    void applyParticleTransformations(const Particle& particle,
                                      GUI::RealSpace::Particles::Particle& particle3D,
                                      const R3& origin = R3(0, 0, 0));

    //! Apply transformations (translation, rotation) to a particle (core/shell) in a
    //! ParticleCoreShell
    void applyParticleCoreShellTransformations(const Particle& particle,
                                               GUI::RealSpace::Particles::Particle& particle3D,
                                               const ParticleCoreShell& particleCoreShell,
                                               const R3& origin = R3(0, 0, 0));

    // Apply color to a 3D particle
    void applyParticleColor(const Particle& particle,
                            GUI::RealSpace::Particles::Particle& particle3D, double alpha = 1);

    std::vector<Particle3DContainer> particle3DContainerVector(const ParticleLayoutItem& layoutItem,
                                                               const QVector3D& origin = {});

    Particle3DContainer singleParticle3DContainer(const Particle& particle,
                                                  double total_abundance = 1.0,
                                                  const QVector3D& origin = {});

    Particle3DContainer particleCoreShell3DContainer(const ParticleCoreShell& particleCoreShell,
                                                     double total_abundance = 1.0,
                                                     const QVector3D& origin = {});

    Particle3DContainer
    particleComposition3DContainer(const ParticleComposition& particleComposition3DContainer,
                                   double total_abundance = 1.0, const QVector3D& origin = {});

    Particle3DContainer mesoCrystal3DContainer(const MesoCrystalItem& mesoCrystalItem,
                                               double total_abundance = 1.0,
                                               const QVector3D& origin = {});

private:
    std::function<QColor(const QString&)> m_fnColorFromMaterialName;
};

} // namespace GUI::RealSpace

#endif // BORNAGAIN_GUI_VIEW_REALSPACE_REALSPACEBUILDERUTILS_H
