//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Slice/ZLimits.h
//! @brief     Defines class ZLimits.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_RESAMPLE_SLICE_ZLIMITS_H
#define BORNAGAIN_RESAMPLE_SLICE_ZLIMITS_H

#include <cmath>
#include <iostream>
#include <limits>

//! An interval. Limits are of type double, and may be infinite.
//! Used for the z-coordinate, especially when slicing form factors.
//!
//! @ingroup intern

class ZLimits {
public:
    ZLimits();
    ZLimits(double _min, double _max);

    bool isFinite() const;

    double zBottom() const { return m_zmin; }
    double zTop() const { return m_zmax; }
    double zTopOr0() const { return std::isfinite(m_zmax) ? m_zmax : 0; }
    double thickness() const { return m_zmax - m_zmin; }
    double thicknessOr0() const { return isFinite() ? m_zmax - m_zmin : 0; }

    static constexpr double inf = std::numeric_limits<double>::infinity();

    //! Returns the union of two ZLimits (the minimal interval that contains both input intervals).
    static ZLimits enclosingInterval(const ZLimits& left, const ZLimits& right);

private:
    // members are not const because we need the implicit copy assignement operator
    double m_zmin;
    double m_zmax;
};

bool operator==(const ZLimits& left, const ZLimits& right);
bool operator!=(const ZLimits& left, const ZLimits& right);

std::ostream& operator<<(std::ostream& ostr, const ZLimits& limits);

#endif // BORNAGAIN_RESAMPLE_SLICE_ZLIMITS_H
#endif // USER_API
