//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Aggregate/InterferenceFinite3DLattice.cpp
//! @brief     Implements class InterferenceFinite2DLattice.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Aggregate/InterferenceFinite3DLattice.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"

#include <limits>

InterferenceFinite3DLattice::InterferenceFinite3DLattice(const Lattice3D& lattice, unsigned N_1,
                                                         unsigned N_2, unsigned N_3)
    : IInterference(0)
    , m_N_1(N_1)
    , m_N_2(N_2)
    , m_N_3(N_3)
{
    setLattice(lattice);
}

InterferenceFinite3DLattice::~InterferenceFinite3DLattice() = default;

InterferenceFinite3DLattice* InterferenceFinite3DLattice::clone() const
{
    auto* result = new InterferenceFinite3DLattice(*m_lattice, m_N_1, m_N_2, m_N_3);
    result->setPositionVariance(m_position_var);
    return result;
}

const Lattice3D& InterferenceFinite3DLattice::lattice() const
{
    if (!m_lattice)
        throw std::runtime_error("InterferenceFinite3DLattice::lattice() -> Error. "
                                 "No lattice defined.");
    return *m_lattice;
}

std::vector<const INode*> InterferenceFinite3DLattice::nodeChildren() const
{
    return std::vector<const INode*>() << m_lattice;
}

double InterferenceFinite3DLattice::iff_without_dw(const R3 q) const
{
    using Math::Laue;
    const double qadiv2 = q.dot(m_lattice->basisVectorA()) / 2.0;
    const double qbdiv2 = q.dot(m_lattice->basisVectorB()) / 2.0;
    const double qcdiv2 = q.dot(m_lattice->basisVectorC()) / 2.0;
    const double ampl = Laue(qadiv2, m_N_1) * Laue(qbdiv2, m_N_2) * Laue(qcdiv2, m_N_3);
    return ampl * ampl / (m_N_1 * m_N_2 * m_N_3);
}

void InterferenceFinite3DLattice::setLattice(const Lattice3D& lattice)
{
    m_lattice = std::make_unique<Lattice3D>(lattice);
}
