//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/SawtoothRipple.h
//! @brief     Defines classes SawtoothRipple*.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_HARDPARTICLE_SAWTOOTHRIPPLE_H
#define BORNAGAIN_SAMPLE_HARDPARTICLE_SAWTOOTHRIPPLE_H

#include "Sample/HardParticle/IProfileRipple.h"

//! The form factor for a cosine ripple, with box profile in elongation direction.
//! @ingroup legacyGrating
class SawtoothRippleBox : public ISawtoothRipple {
public:
    SawtoothRippleBox(double length, double width, double height, double asymmetry);
#ifndef USER_API
    SawtoothRippleBox(std::vector<double> P);
    SawtoothRippleBox* clone() const override;
    std::string className() const final { return "SawtoothRippleBox"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Length", "nm", "Characteristic length", 0, INF, 1.},
                {"Width", "nm", "Width", 0, INF, 1.},
                {"Height", "nm", "Height", 0, INF, 1.},
                {"AsymmetryLength", "nm", "Asymmetry of width", -INF, INF, 0.}};
    }

private:
    complex_t factor_x(complex_t qx) const override;
#endif // USER_API
};

//! The form factor for a cosine ripple, with Gaussian profile in elongation direction.
//! @ingroup legacyGrating
class SawtoothRippleGauss : public ISawtoothRipple {
public:
    SawtoothRippleGauss(double length, double width, double height, double asymmetry);
#ifndef USER_API
    SawtoothRippleGauss(std::vector<double> P);
    SawtoothRippleGauss* clone() const override;
    std::string className() const final { return "SawtoothRippleGauss"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Length", "nm", "Characteristic length", 0, INF, 1.},
                {"Width", "nm", "Width", 0, INF, 1.},
                {"Height", "nm", "Height", 0, INF, 1.},
                {"AsymmetryLength", "nm", "Asymmetry of width", -INF, INF, 0.}};
    }

private:
    complex_t factor_x(complex_t qx) const override;
#endif // USER_API
};

//! The form factor for a cosine ripple, with Lorentz form factor in elongation direction.
//! @ingroup legacyGrating
class SawtoothRippleLorentz : public ISawtoothRipple {
public:
    SawtoothRippleLorentz(double length, double width, double height, double asymmetry);
#ifndef USER_API
    SawtoothRippleLorentz(std::vector<double> P);
    SawtoothRippleLorentz* clone() const override;
    std::string className() const final { return "SawtoothRippleLorentz"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Length", "nm", "Characteristic length", 0, INF, 1.},
                {"Width", "nm", "Width", 0, INF, 1.},
                {"Height", "nm", "Height", 0, INF, 1.},
                {"AsymmetryLength", "nm", "Asymmetry of width", -INF, INF, 0.}};
    }

private:
    complex_t factor_x(complex_t qx) const override;
#endif // USER_API
};

#endif // BORNAGAIN_SAMPLE_HARDPARTICLE_SAWTOOTHRIPPLE_H
