//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Tests/Functional/Core/SuitePersist/Check.cpp
//! @brief     Implements function checkSimulation for core standard test
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "BABuild.h"
#include "BATesting.h"
#include "Base/Util/FileSystemUtils.h"
#include "Device/Histo/HistoUtils.h"
#include "Device/Histo/IOFactory.h"
#include "Device/Histo/SimulationResult.h"
#include "Sim/Simulation/ISimulation.h"
#include <iostream>

//! Run simulation directly (in C+ core), and compare result with reference data.

bool checkSimulation(const std::string& name, ISimulation& direct_simulation, const double limit)
{
    // Run simulation directly.
    const std::unique_ptr<Powerfield<double>> result_data = direct_simulation.simulate().data();

    std::unique_ptr<Powerfield<double>> reference;

    // Load reference if available.
    ASSERT(!name.empty());
    try {
        const std::string refPath =
            BaseUtils::Filesystem::jointPath(BATesting::ReferenceDir_Suite(), name + ".int.gz");
        std::cout << "- reference: " << refPath << "\n";
        reference.reset(IOFactory::readPowerfield(refPath));
    } catch (const std::exception&) {
        std::cerr << "FAILED: cannot read reference\n";
    }

    // Compare with reference if available.
    if (reference) {
        std::cout << "- check diff" << std::endl;
        if (DataUtils::Histo::checkRelativeDifference(*result_data, *reference, limit)) {
            std::cout << "- success" << std::endl;
            return true; // regular exit
        }
    } else {
        std::cerr << "FAILED: reference not found\n";
    }

    // Save simulation, as it differs from reference.
    BaseUtils::Filesystem::createDirectories(BATesting::TestOutDir_Suite());
    std::string out_fname =
        BaseUtils::Filesystem::jointPath(BATesting::TestOutDir_Suite(), name + ".int.gz");
    IOFactory::writePowerfield(*result_data, out_fname);
    std::cout << "Notes:\n- to visualize an intensity map, use " << BABuild::srcDir()
              << "/devtools/view/plot_int.py\n"
              << "- to plot a difference image, use " << BABuild::srcDir()
              << "/devtools/view/plot_diff_int.py\n"
              << "- if the new simulation result\n"
              << "    " << out_fname << "\n"
              << "  is correct, then move it to the reference directory\n"
              << "    " << BATesting::ReferenceDir_Suite() << "/\n\n";

    return false;
}
