#include "GUI/Model/BaseItem/PropertyItem.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/RectangularDetectorItem.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Model/Model/SessionXML.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Support/XML/Backup.h"
#include "GUI/Support/XML/Streamer.h"
#include "Tests/GTestWrapper/google_test.h"
#include <QXmlStreamReader>
#include <QXmlStreamWriter>
#include <memory>

namespace {

QString itemToXML(SessionItem* item)
{
    QString result;
    QXmlStreamWriter writer(&result);
    GUI::Session::XML::writeModel(&writer, item);
    return result;
}

void itemFromXML(QString buffer, SessionItem* item)
{
    QXmlStreamReader reader(buffer);
    GUI::Session::XML::readItems(&reader, item);
}

template <typename ItemType>
QByteArray serialize(ItemType& t)
{
    return GUI::Util::createBackup(&t);
}

template <typename ItemType>
void deserialize(ItemType& t, const QByteArray& a)
{
    GUI::Util::restoreBackup(&t, a);
}

} // namespace

class TestSessionXML : public ::testing::Test {
};

//! Testing to/from xml: simple property item.

TEST_F(TestSessionXML, sessionItem)
{
    QString expected;

    SessionModel source("TestModel");
    source.insertItem<PropertyItem>();

    expected = "<TestModel>"
               "<Item ModelType=\"Property\" Tag=\"rootTag\" DisplayName=\"Property\"/>"
               "</TestModel>";
    EXPECT_EQ(itemToXML(source.rootItem()), expected);

    SessionModel target("TestModel");
    itemFromXML(expected, target.rootItem());

    EXPECT_EQ(target.rowCount(QModelIndex()), 1);
    SessionItem* newItem = target.itemForIndex(target.index(0, 0, QModelIndex()));
    EXPECT_TRUE(newItem->hasModelType<PropertyItem>());
    EXPECT_EQ(newItem->displayName(), "Property");
    EXPECT_FALSE(newItem->value().isValid());
}

//! Testing to/from xml: FullSphereItem

TEST_F(TestSessionXML, InstrumentItem)
{
    SpecularInstrumentItem instrument;
    instrument.setBackgroundType<ConstantBackgroundItem>();
    const auto a1 = serialize(instrument);

    // target model, from xml
    SpecularInstrumentItem target;
    deserialize(target, a1);

    // checking top items in source and target models
    auto* backgroundItem = instrument.backgroundItem();

    EXPECT_EQ(instrument.instrumentName(), target.instrumentName());
    EXPECT_EQ(instrument.id(), target.id());
    EXPECT_EQ(instrument.analyzerEfficiency(), target.analyzerEfficiency());

    BackgroundItem* t_radius = target.backgroundItem();
    EXPECT_TRUE(dynamic_cast<ConstantBackgroundItem*>(backgroundItem));
    EXPECT_TRUE(dynamic_cast<ConstantBackgroundItem*>(t_radius));
    EXPECT_EQ(dynamic_cast<ConstantBackgroundItem*>(backgroundItem)->backgroundValue().get(),
              dynamic_cast<ConstantBackgroundItem*>(t_radius)->backgroundValue().get());

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, GISASInstrumentItem)
{
    GISASInstrumentItem instrument;
    const auto a1 = serialize(instrument);

    GISASInstrumentItem target;
    deserialize(target, a1);

    // checking top items in source and target models
    EXPECT_EQ(instrument.instrumentName(), target.instrumentName());
    EXPECT_EQ(instrument.id(), target.id());
    EXPECT_EQ(instrument.analyzerEfficiency(), target.analyzerEfficiency());

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, OffSpecularInstrumentItem)
{
    OffSpecularInstrumentItem instrument;
    const auto a1 = serialize(instrument);

    OffSpecularInstrumentItem target;
    deserialize(target, a1);

    // checking top items in source and target models
    EXPECT_EQ(instrument.instrumentName(), target.instrumentName());
    EXPECT_EQ(instrument.id(), target.id());
    EXPECT_EQ(instrument.analyzerEfficiency(), target.analyzerEfficiency());

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, SpecularInstrumentItem)
{
    SpecularInstrumentItem instrument;
    const auto a1 = serialize(instrument);

    SpecularInstrumentItem target;
    deserialize(target, a1);

    // checking top items in source and target models
    EXPECT_EQ(instrument.instrumentName(), target.instrumentName());
    EXPECT_EQ(instrument.id(), target.id());
    EXPECT_EQ(instrument.analyzerEfficiency(), target.analyzerEfficiency());

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, DepthProbeInstrumentItem)
{
    DepthProbeInstrumentItem instrument;
    const auto a1 = serialize(instrument);

    DepthProbeInstrumentItem target;
    deserialize(target, a1);

    // checking top items in source and target models
    EXPECT_EQ(instrument.instrumentName(), target.instrumentName());
    EXPECT_EQ(instrument.id(), target.id());
    EXPECT_EQ(instrument.analyzerEfficiency(), target.analyzerEfficiency());
    EXPECT_EQ(instrument.zAxis().min(), target.zAxis().min());

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}


TEST_F(TestSessionXML, emptyMultiLayer)
{
    MultiLayerItem source;
    const auto a1 = serialize(source);

    MultiLayerItem target;
    deserialize(target, a1);

    // final XML comparison
    const auto a2 = serialize(target);
    EXPECT_EQ(a1, a2);
}

TEST_F(TestSessionXML, Layer)
{
    MaterialItems materials;
    LayerItem source(&materials);
    const auto a1 = serialize(source);

    LayerItem target(&materials);
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}

TEST_F(TestSessionXML, Particle)
{
    MaterialItems materials;
    ParticleItem source(&materials);
    const auto a1 = serialize(source);

    ParticleItem target(&materials);
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}

TEST_F(TestSessionXML, ParticleWithFF)
{
    MaterialItems materials;
    ParticleItem source(&materials);
    source.setFormFactorType<Pyramid2Item>();
    const auto a1 = serialize(source);

    ParticleItem target(&materials);
    deserialize(target, a1);

    // final XML comparison
    EXPECT_EQ(a1, serialize(target));
}

TEST_F(TestSessionXML, Alignment)
{
    RectangularDetectorItem r;
    r.setDetectorAlignment(RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM_DPOS);
    const auto a1 = serialize(r);
    r.setDetectorAlignment(RectangularDetector::GENERIC);
    deserialize(r, a1);
    EXPECT_EQ(r.detectorAlignment(), RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM_DPOS);
}